"""
This module provides helper functions for working with images.

Functions:
    get_supported_image_formats:
        Get the supported image formats.
"""

from pathlib import Path
from typing import Union

from platformdirs import user_pictures_path

from inspyre_toolbox.filesystem.file.collection import FileCollection, collect_files
from . import VALID_IMAGE_EXTENSIONS

DEFAULT_PICTURES_DIR = user_pictures_path()


def collect_image_files(root_dir: Union[str, Path] = DEFAULT_PICTURES_DIR, recursive: bool = True,
                        file_types=VALID_IMAGE_EXTENSIONS, **kwargs) -> FileCollection:
    return collect_files(root_dir, recursive, file_types=file_types, **kwargs)



import hashlib
from typing import Union

from inspyre_toolbox.conversions.bytes import ByteConverter


def get_file_checksum(file_path, algorithm='sha256'):
    """
    Calculate the checksum of a file using the specified hashing algorithm.

    Parameters:
    file_path (str): Path to the file.
    algorithm (str): Hashing algorithm to use ('md5', 'sha1', 'sha256', etc.). Default is 'sha256'.

    Returns:
    str: The calculated checksum.

    Raises:
    ValueError: If the specified algorithm is not supported.
    FileNotFoundError: If the specified file does not exist.
    """
    # Validate the algorithm
    if algorithm not in hashlib.algorithms_available:
        raise ValueError(f"Unsupported algorithm '{algorithm}'. Available algorithms: {hashlib.algorithms_available}")

    # Initialize the hash object
    hash_func = hashlib.new(algorithm)

    # Read the file in chunks and update the hash object
    try:
        with open(file_path, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_func.update(chunk)
    except FileNotFoundError:
        raise FileNotFoundError(f"The file '{file_path}' does not exist.")

    # Return the hex digest of the hash
    return hash_func.hexdigest()


def get_lowest_unit_size(size: int) -> tuple[Union[int, float], str]:
    """
    Get the lowest unit size for a given size.

    Parameters:
        size (int):
            The size to convert.

    Returns:
        tuple[Union[int, float], str]:
            The converted size and the unit.

    Examples:
        >>> get_lowest_unit_size(1024)
        (1.0, 'KB')
        >>> get_lowest_unit_size(1024 * 1024)
        (1.0, 'MB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024)
        (1.0, 'GB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024)
        (1.0, 'TB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'PB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'EB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'ZB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'YB')
    """
    units = ['byte', 'kilobyte', 'megabyte', 'gigabyte', 'terabyte', 'petabyte', 'exabyte', 'zetabyte', 'yottabyte']
    units.reverse()
    converter = ByteConverter(size, 'byte')

    for unit in units:
        converted = converter.convert(unit.lower())

        if converted >= 1:
            return converted, unit.upper(),



from typing import Union

from inspy_logger import InspyLogger, Loggable

from inspyre_toolbox.humanize import Numerical
from inspyre_toolbox.syntactic_sweets.classes import validate_type
from inspyre_toolbox.version import parse_version

VERSION = parse_version()

MOD_LOGGER = InspyLogger('inspyre_toolbox.conversions.bytes')

MOD_LOGGER.debug(f'Byte conversion module loaded. | Inspyre-Toolbox v{VERSION}')


class ByteConverter(Loggable):
    """
    A class to convert values between different byte units.

    Attributes:
        units (dict):
            A dictionary holding the conversion units.

        BYTE_FAMILY (list):
            A list of byte units.

        BIT_FAMILY (list):
            A list of bit units.

    Methods:
        convert(unit: str) -> float:
            Converts the bytes value to the specified unit.

        get_lowest_safe_conversion(keep_family: bool) -> tuple(float, str):
            Returns the lowest safe conversion for the value, a tuple containing the value and unit,
            respectively.
    """
    # Dictionary holding the conversion units
    UNITS = {
        'byte': 1,
        'kilobyte': 1024,
        'megabyte': 1024**2,
        'gigabyte': 1024**3,
        'terabyte': 1024**4,
        'petabyte': 1024**5,
        'exabyte': 1024**6,
        'zetabyte': 1024**7,
        'yottabyte': 1024**8,
        'bit': 1/8,
        'kilobit': 1024/8,
        'megabit': (1024**2)/8,
        'gigabit': (1024**3)/8,
        'terabit': (1024**4)/8,
        'petabit': (1024**5)/8,
        'exabit': (1024**6)/8,
        'zetabit': (1024**7)/8,
        'yottabit': (1024**8)/8,
            }

    BYTE_FAMILY = [
            'byte', 'kilobyte', 'megabyte',
            'gigabyte', 'terabyte', 'petabyte',
            'exabyte', 'zetabyte', 'yottabyte'
            ]

    BIT_FAMILY = [
            'bit', 'kilobit', 'megabit',
            'gigabit', 'terabit', 'petabit',
            'exabit', 'zetabit', 'yottabit'
            ]

    def __init__(
            self,
            value: Union[int, float],
            unit: str,
            keep_family: bool = False
            ):
        """
        Initializes the ByteConverter class.

        Parameters:
            value (Union[int, float]):
                The value to convert.

            unit (str):
                The unit of the value.

            keep_family (bool):
                A flag indicating whether to keep the family of the initial unit. To keep the
                "family" of a unit means to keep the units that are in the same family as the
                initial unit.

                Examples:
                     - The initial unit is 'kilobyte';
                         The family would be
                            - kilobyte
                            - megabyte
                            - gigabyte
                            - terabyte
                            - petabyte
                            - exabyte
                            - zetabyte
                            - yottabyte

                    - The initial unit is 'kilobit';
                        The family would be
                            - kilobit
                            - megabit
                            - gigabit
                            - terabit
                            - petabit
                            - exabit
                            - zetabit
                            - yottabit
        """
        super().__init__(parent_log_device=MOD_LOGGER)
        self.__keep_family = None

        unit = unit.removesuffix('s')

        if unit not in self.UNITS:
            raise ValueError(f'Invalid unit: {unit}')

        if value < 0:
            raise ValueError('Value cannot be negative.')

        self.__initial_unit = unit
        self.__initial_value = value

        # Convert the input value to bytes
        self.__bytes = value * self.UNITS[unit]
        self.keep_family = keep_family

        self.__family_factors = [self.UNITS[u] for u in self.UNIT_FAMILY]

    def convert(self, unit: str, strict_case: bool = False):
        """
        Converts the bytes value to the specified unit.

        Parameters:
            unit (str):
                The unit to convert the value to. (Must be a key in the :ref:`units` dictionary)

            strict_case (bool):
                A flag indicating whether to convert the unit to lowercase before checking if it is
                a valid unit.

        Returns:
            float:
                The converted value.
        """
        log_name = f'{self.log_device.name}:convert'
        log = self.log_device.find_child_by_name(log_name)[0] if self.log_device.has_child(
                log_name) else self.create_child_logger()

        unit = unit.removesuffix('s')

        if not strict_case:
            log.debug(f'Converting case ({unit}) to lower.')
            unit = unit.lower()

        log.debug(f'Searching valid units for {unit}...')
        if unit not in self.UNITS:
            log.error(f'Invalid unit: {unit}')
            raise ValueError(f'Invalid unit: {unit}')

        # Convert the bytes value to the specified unit
        return self.initial_bytes / self.units[unit]

    def get_lowest_safe_conversion(
            self,
            keep_family: bool = False
            ) -> tuple[float, str]:
        """
        Returns the lowest safe conversion for the value.

        Parameters:
            keep_family (bool):
                A flag indicating whether to keep the family of the initial unit.

        Returns:
            tuple:
                The lowest safe conversion value and unit.
        """
        log_name = f'{self.log_device.name}:get_lowest_safe_conversion'
        log = self.log_device.find_child_by_name(log_name)[0] if self.log_device.has_child(
            log_name) else self.create_child_logger()

        if keep_family:
            log.debug(f'Keeping the family of the initial unit: {self.__initial_unit}')
            family = self.BYTE_FAMILY if self.__initial_unit in self.BYTE_FAMILY else self.BIT_FAMILY
            log.debug(f'Family of the initial unit: {family}')
            factors = [self.UNITS[unit] for unit in family]
        else:
            log.debug('Not keeping the family of the initial unit.')
            factors = [self.UNITS[unit] for unit in self.UNITS]

        log.debug(f'Factors: {factors}')

        base_value = self.__initial_value * self.UNITS[self.__initial_unit]
        log.debug(f'Base value: {base_value}')

        lowest_value = float('inf')
        lowest_unit = None

        for unit, factor in self.UNITS.items():
            log.debug(f'Checking unit: {unit}')
            log.debug(f'Factor: {factor}')
            if factor in factors:
                log.debug(f'Factor in factors: {factor}')
                converted_value = base_value / factor
                log.debug(f'Converted value: {Numerical(converted_value, unit).count_noun()}')
                if 1.0 <= converted_value < lowest_value:
                    log.debug(f'Converted value is less than lowest value: {converted_value}')
                    lowest_value = converted_value
                    log.debug(f'New lowest value: {lowest_value}')
                    lowest_unit = unit
                    log.debug(f'New lowest unit: {lowest_unit}')

        log.debug(f'Lowest value: {lowest_value}, Lowest unit: {lowest_unit}')

        if lowest_unit is None:
            log.debug('No unit found. Finding the smallest factor.')
            smallest_factor_index = factors.index(min(factors))
            log.debug(f'Smallest factor index: {smallest_factor_index}')
            lowest_unit = list(self.UNITS.keys())[smallest_factor_index]
            log.debug(f'Lowest unit: {lowest_unit}')
            lowest_value = base_value / min(factors)
            log.debug(f'Lowest value: {Numerical(lowest_value, lowest_unit).count_noun()}')

        log.debug(f'Final lowest value: {lowest_value}, Final lowest unit: {lowest_unit} | '
                  f'{Numerical(lowest_value, lowest_unit).count_noun()}')

        return lowest_value, lowest_unit

    @property
    def conversion_map(self):
        """
        Returns the conversion map.

        This method returns a dictionary containing the conversion units based on the value of the
        keep_family attribute.

        Returns:
            dict:
                The conversion map.

        """
        return {unit: self.UNITS[unit] for unit in (self.UNIT_FAMILY if self.keep_family else self.UNITS)}

    @property
    def family_factors(self) -> list[float]:
        """
        Returns the factors of the family of the initial unit.

        Returns:
            list:
                The factors of the family of the initial unit.

        """
        return self.__family_factors

    @property
    def keep_family(self):
        """
        Returns the keep_family attribute.

        Returns:
            bool:
                The keep_family attribute.
        """
        return self.__keep_family

    @keep_family.setter
    @validate_type(bool)
    def keep_family(self, new):
        """
        Sets the keep_family attribute.

        Parameters:
            new (bool):
                The value to set the keep_family attribute to.

        Returns:
            None
        """
        self.__keep_family = new

    @property
    def initial_bytes(self):
        """
        Returns the initial value in bytes.

        Returns:
            float:
                The initial value in bytes.
        """
        return self.__bytes

    @property
    def initial_unit(self) -> str:
        """
        Returns the initial unit of the value.

        Returns:
            str:
                The initial unit of the value
        """
        return self.__initial_unit

    @initial_unit.setter
    def initial_unit(self, unit):
        """
        Placeholder for the initial unit setter.

        This setter is a placeholder and raises an AttributeError when called.

        Parameters:
            unit:
                Unused parameter.

        Returns:
            None

        Raises:
            AttributeError:
                The initial unit cannot be changed.
        """
        raise AttributeError("The initial unit cannot be changed.")

    @property
    def initial_value(self):
        """
        Returns the initial value.

        Returns:
            Union[int, float]:
                The initial value.
        """
        return self.__initial_value

    @initial_value.setter
    def initial_value(self, value):
        """
        Placeholder for the initial value setter.

        This setter is a placeholder and raises an AttributeError when called.

        Parameters:
            value:
                Unused parameter.

        Returns:
            None

        Raises:
            AttributeError:
                The initial value cannot be changed.
        """
        raise AttributeError("The initial value cannot be changed.")

    @property
    def units(self):
        return self.UNITS

    @property
    def UNIT_FAMILY(self):
        """
        Returns the family of the initial unit.

        Returns:
            list:
                The family of the initial unit.
        """
        if not self.initial_unit:
            return None

        return self.BYTE_FAMILY if self.initial_unit in self.BYTE_FAMILY else self.BIT_FAMILY

    @property
    def UNIT_FAMILY_NAME(self):
        """
        Returns the name of the family of the initial unit.

        Returns:
            str:
                The name of the family of the initial unit.
        """
        if not self.initial_unit:
            return None

        return 'byte' if self.initial_unit in self.BYTE_FAMILY else 'bit'

    # Below are property methods that return the value in a specific unit

    @property
    def bits(self):
        """Returns the value in bits"""
        return self.convert('bit')

    @property
    def kilobits(self):
        """Returns the value in kilobits"""
        return self.convert('kilobit')

    @property
    def megabits(self):
        """Returns the value in megabits"""
        return self.convert('megabit')

    @property
    def gigabits(self):
        """Returns the value in gigabits"""
        return self.convert('gigabit')

    @property
    def terabits(self):
        """Returns the value in terabits"""
        return self.convert('terabit')

    @property
    def petabits(self):
        """Returns the value in petabits"""
        return self.convert('petabit')

    @property
    def exabits(self):
        """Returns the value in exabits"""
        return self.convert('exabit')

    @property
    def zetabits(self):
        """Returns the value in zetabits"""
        return self.convert('zetabit')

    @property
    def yottabits(self):
        """Returns the value in yottabits"""
        return self.convert('yottabit')

    @property
    def bytes(self):
        """Returns the value in bytes"""
        return self.convert('byte')

    @property
    def kilobytes(self):
        """Returns the value in kilobytes"""
        return self.convert('kilobyte')

    @property
    def megabytes(self):
        """Returns the value in megabytes"""
        return self.convert('megabyte')

    @property
    def gigabytes(self):
        """Returns the value in gigabytes"""
        return self.convert('gigabyte')

    @property
    def terabytes(self):
        """Returns the value in terabytes"""
        return self.convert('terabyte')

    @property
    def petabytes(self):
        """Returns the value in petabytes"""
        return self.convert('petabyte')

    @property
    def exabytes(self):
        """Returns the value in exabytes"""
        return self.convert('exabyte')

    @property
    def zetabytes(self):
        """Returns the value in zetabytes"""
        return self.convert('zetabyte')

    @property
    def yottabytes(self):
        """Returns the value in yottabytes"""
        return self.convert('yottabyte')

    def __str__(self):
        return f'{self.__initial_value} {self.__initial_unit}'

    def __repr__(self):
        return \
            (
                    f'<ByteConverter@{hex(id(self))}: '
                    f'Initial value: {self.__initial_value} | Initial unit: {self.__initial_unit}>'
            )


def get_lowest_unit_size(size: int) -> tuple[Union[int, float], str]:
    """
    Get the lowest unit size for a given size.

    Parameters:
        size (int):
            The size to convert.

    Returns:
        tuple[Union[int, float], str]:
            The converted size and the unit.

    Examples:
        >>> get_lowest_unit_size(1024)
        (1.0, 'KB')
        >>> get_lowest_unit_size(1024 * 1024)
        (1.0, 'MB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024)
        (1.0, 'GB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024)
        (1.0, 'TB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'PB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'EB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'ZB')
        >>> get_lowest_unit_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024)
        (1.0, 'YB')
    """
    units = ['byte', 'kilobyte', 'megabyte', 'gigabyte', 'terabyte', 'petabyte', 'exabyte', 'zetabyte', 'yottabyte']
    units.reverse()
    converter = ByteConverter(size, 'byte')

    for unit in units:
        converted = converter.convert(unit.lower())

        if converted >= 1:
            return converted, unit.upper(),



"""
Since:
    1.6.0
"""
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Union

from box import Box

from inspyre_toolbox.conversions.bytes import get_lowest_unit_size
from inspyre_toolbox.filesystem.errors import NeedsProcessingError
from inspyre_toolbox.filesystem.file import MOD_LOGGER as PARENT_LOGGER
from inspyre_toolbox.humanize import Numerical
from inspyre_toolbox.log_engine import Loggable
from inspyre_toolbox.path_man import gather_files_in_dir, prepare_path, provision_path, provision_paths
from inspyre_toolbox.syntactic_sweets.locks import flag_lock
from . import File

MOD_LOGGER = PARENT_LOGGER.get_child('collection')


class NeedsReprocessingTag:

    def __init__(self, value: bool = False):
        self.__value = value if isinstance(value, bool) else None
        if self.__value is None:
            raise ValueError("Value must be a boolean.")

    def __get__(self, instance, owner):
        return self.__value

    def __set__(self, instance, value):
        if not isinstance(value, bool):
            raise ValueError("Value must be a boolean.")
        self.__value = value


@dataclass
class FileCollection(Loggable):
    """
    A class for managing a collection of files.

    Properties:
        paths (list):
            A list of file paths.

        total_size (int):
            The total size of the collection, in bytes.

        total_files (int):
            The total number of files in the collection.

        extensions (dict):
            A dictionary of extensions with classes for total size and number of files.

        needs_reprocessing (bool):
            A flag indicating whether the collection needs reprocessing.

    Methods:
        get_total_size_in_lowest_unit:
            Get the total size of the collection in the lowest unit.

        get_total_extension_size_in_lowest_unit:
            Get the total size of a specific extension in the lowest unit.

        remove_file:
            Remove a file from the collection.

        reprocess_files:
            Reprocess the files in the collection.

    Examples:
        >>> collection = FileCollection(paths=['/path/to/file1', '/path/to/file2'])
        >>> collection.total_size
        123456  # Total size of the collection, in bytes.
    """
    paths: list
    total_size: int = field(init=False, default=0)
    total_files: int = field(init=False, default=0)
    extensions: dict = field(init=False, default_factory=dict)

    def __init__(self, paths: List[Union[str, Path]] = None, auto_process: bool = False):
        """
        Initialize the FileCollection with a list of file paths.

        `FileCollection` is a class for managing a collection of files. It takes a list of file paths and creates a class
        representing a collection of files with classes for total size and number of files and a dictionary of extensions
        with classes for total size and number of files.

        Parameters:
            paths (list):
                A list of file paths.

        Returns:
            None
        """
        super().__init__(MOD_LOGGER)
        self._files = {
                'local':  {
                        'files':      [],
                        'total_size': 0,
                        },
                'remote': {
                        'files':      [],
                        'total_size': 0,
                        }
                }

        self._files = Box(self._files)
        self._needs_processing = False
        self._files_gathered = False
        self._processing = False
        self.paths = paths

        if self.paths:
            self._files_gathered = True
            self._needs_processing = True

        self.total_size = 0
        self.total_files = 0
        self.extensions = {}
        self.__needs_reprocessing = NeedsReprocessingTag()

        if auto_process:
            self.process_files()

    @property
    def files(self):
        if self.needs_processing and not self._processing:
            raise NeedsProcessingError("Files need to be processed before they can be accessed.")

        return self._files

    @property
    def files_gathered(self) -> bool:
        """
        Returns whether the files have been gathered.

        This property returns a boolean value indicating whether the files have been gathered. If the files have been
        gathered, this property will return `True`. If the files have not been gathered, it will return `False`. The
        property is read-only.

        Returns:
            bool:
                True if the files have been gathered, False otherwise.
        """
        return self._files_gathered

    @property
    def needs_processing(self):
        return self._needs_processing

    @property
    def needs_reprocessing(self) -> bool:
        """
        Returns whether the collection needs reprocessing.

        This property returns a boolean value indicating whether the collection needs reprocessing. If the collection has
        been modified since the last processing, this property will return `True`. If the collection has not been modified,
        it will return `False`. The property is read-only.

        Returns:
            bool:
                True if the collection needs reprocessing, False otherwise.
        """
        return self.__needs_reprocessing

    @needs_reprocessing.setter
    def needs_reprocessing(self, value: bool):
        """
        Sets whether the collection needs reprocessing.

        This method sets the `needs_reprocessing` attribute of the collection. If the value is `True`, the collection will
        be reprocessed the next time it is accessed. If the value is `False`, the collection will not be reprocessed. The
        method is used to indicate that the collection has been modified and needs to be reprocessed.

        Parameters:
            value (bool):
                The value to set the `needs_reprocessing` attribute to. True if the collection needs reprocessing, False
                otherwise.

        Returns:
            None
        """
        self.__needs_reprocessing = value

    @property
    def total_local_size(self):
        return self.files['local']['total_size']

    @property
    def total_remote_size(self):
        return self.files['remote']['total_size']

    def process_files(self):
        """
        Process the files in the collection.

        This method processes the files in the collection. It calculates the total size of the collection, the total number
        of files, and the total size of each extension in the collection. It populates the `total_size`, `total_files`, and
        `extensions` attributes of the class.

        Returns:
            None
        """
        self.paths = provision_paths(self.paths)

        with flag_lock(self, 'processing'):

            for path_str in self.paths:
                file = File(path_str)
                if file.has_recall_attribute:
                    bucket = self.files['remote']
                else:
                    bucket = self.files['local']

                bucket['files'].append(file)
                bucket['total_size'] += file.size_in_bytes

                if file.extension not in self.extensions:
                    self.extensions[file.extension] = {
                            'total_size':  0,
                            'total_files': 0,
                            }

                self.total_size += file.size_in_bytes
                self.total_files += 1

                self.extensions[file.extension]['total_size'] += file.size_in_bytes
                self.extensions[file.extension]['total_files'] += 1

        self._needs_processing = False

    def reprocess_files(self):
        """
        Reprocess the files in the collection.

        This method reprocesses the files in the collection. It recalculates the total size of the collection, the total
        number of files, and the total size of each extension in the collection. It populates the `total_size`, `total_files`,
        and `extensions` attributes of the class. This method is used to update the collection after files have been added
        or removed.

        Note:
            This method should be called after files have been added or removed from the collection. It will only reprocess
            the files if the `needs_reprocessing` attribute is set to `True`. After reprocessing, the `needs_reprocessing`
            attribute will be set to `False`.

        Returns:
            None

        """
        if self.needs_reprocessing:
            self.total_size = 0
            self.total_files = 0
            self.extensions = {}
            self.process_files()

    def find_file(self, path: Union[str, Path], include_remote: bool = False) -> Union[File, None]:
        """
        Find a file in the collection.

        This method finds a file in the collection by its path. It takes a file path as an argument and returns the file
        object if the file is in the collection. If the file is not in the collection, the method will return `None`.

        Parameters:
            path (Union[str, Path]):
                The file path to find.

            include_remote (bool):
                A flag indicating whether to include remote files in the search. If True, the method will search both
                local and remote files. If False, the method will only search local files.

        Returns:
            Union[File, None]:
                The file object if the file is in the collection, `None` otherwise.
        """
        log = self.create_logger()

        log.debug(f"Finding file: {path}")
        path = provision_path(path)

        log.debug(f'Searching {len(self.files["local"]["files"])} local files....')

        local_acc = None

        for local_acc, file in enumerate(self.files['local']['files']):
            if file.path == path:
                log.debug(f"Found file: {file.path} after {local_acc} iterations.")
                return file

        log.debug(f'File not found after searching {local_acc} files.')

        if include_remote:
            log.debug(f'Searching {len(self.files["remote"]["files"])} remote files....')
            for remote_acc, file in enumerate(self.files['remote']['files']):
                if file.path == path:
                    log.debug(f"Found file: {file.path} after {local_acc + remote_acc} iterations.")
                    return file

        return None

    def get_total_size_in_lowest_unit(self) -> tuple[Union[int, float], str]:
        """
        Get the total size of the collection in the lowest unit with a size greater than or equal to 1.

        This method calculates the total size of the collection in the lowest unit with a size greater than or equal
        to 1. It returns the size and the unit as a tuple. The unit is a string representing the unit of the size.

        Returns:
            tuple[Union[int, float], str]:
                The total size of the collection and the unit.

        """
        return get_lowest_unit_size(self.total_size)

    def get_total_extension_size_in_lowest_unit(
            self,
            extension: str,
            return_as_string: bool = False
            ) -> tuple[Union[int, float], str]:
        """
        Get the total size of a specific extension in the lowest unit with a size greater than or equal to 1.

        Parameters:
            extension (str):
                The extension to get the size of.

            return_as_string (bool):
                A flag indicating whether to return the size as a string. If True, the size will be returned as a string
                with the unit. If False, the size will be returned as a tuple with the size and the unit.

        Returns:
            tuple[Union[int, float], str]:
                The total size of the extension and the unit.

        Raises:
            ValueError:
                If the extension is not found in the collection.
        """

        if not extension.startswith('.'):
            extension = f".{extension}"

        extension = extension.lower()

        if extension not in self.extensions:
            raise ValueError(f"Invalid extension: {extension}!")

        lowest_unit_size = get_lowest_unit_size(self.extensions[extension]['total_size'])

        if return_as_string:
            return Numerical(lowest_unit_size[0], noun=lowest_unit_size[1]).count_noun()
        return lowest_unit_size

    def remove_file(self, path: Union[Path, str], **kwargs):
        """
        Remove a file from the collection.

        This method removes a file from the collection. It takes a file path as an argument and removes the file from
        the collection. It then reprocesses the files in the collection to update the total size, total number of files,
        and total size of each extension.

        Note: This method does not delete the file from the file system. It only removes the file from the
        collection, and only if the file is in the collection. If the file is not in the collection, this method will
        raise a `KeyError`. After removing the file, the collection will be reprocessed to update the total size,
        total number of files, and total size of each extension.

        Parameters:
            path (Union[Path, str]):
                The file path to remove.

            **kwargs:
                Additional keyword arguments.

        Raises:
            KeyError:
                If the file is not in the collection.

        Returns:
            None
        """
        path = provision_path(path, **kwargs)
        if path in self.paths:
            self.paths.remove(path)
            self.needs_reprocessing = True
            self.reprocess_files()
            self.needs_reprocessing = False

    def __getitem__(self, key: Union[int, str]) -> Union[str, Dict[str, int]]:
        if isinstance(key, int):
            return str(self.paths[key])
        elif isinstance(key, str):
            if not key.startswith('.'):
                key = f".{key}"
            key = key.lower()
            if key not in self.extensions:
                raise KeyError(f"Extension {key} not found in file collection.")
            return self.extensions[key]
        else:
            raise TypeError("Key must be an integer (for path index) or a string (for extension).")

    def __str__(self):
        size, unit = self.get_total_size_in_lowest_unit()
        size_str = Numerical(size, noun=unit).count_noun()
        return f"Total size: {size_str}, Total files: {self.total_files}"


def create_file_collection(paths: list[Path]) -> FileCollection:
    """
    Create a file collection.

    This function creates a file collection from a list of file paths. It takes a list of file paths as an argument and
    returns a :class:`FileCollection` object representing the collection of files.

    Parameters:
        paths (list[Path]):
            A list of file paths.

    Returns:
        FileCollection:
            A `FileCollection` object representing the collection of files.

    Since:
        v1.0.0
    """
    fc = FileCollection(paths)
    fc.process_files()
    return fc


def collect_files(root_dir: Union[str, Path], recursive=True, **kwargs):
    """
    Collect files in a directory into a :class:`FileCollection`.

    Parameters:
        root_dir (Union[str, Path]):
            The root directory from which to collect files.

        recursive (bool):
            A flag indicating whether to collect files recursively.

        **kwargs:
            Additional keyword arguments.

    Returns:
        FileCollection:
            A `FileCollection` object representing the collection of files.

    """
    root_dir = prepare_path(root_dir, do_not_create=True)
    files = gather_files_in_dir(root_dir, recursive=recursive, **kwargs)
    return create_file_collection(files)



from inspyre_toolbox.common.meta import RELEASE_MAP, VERSION as _VERSION
from inspyre_toolbox.ver_man import PyPiVersionInfo


def get_full_version_name():
    """
    Gets the full version name.

    Returns:
        str: The full version name.

    Since:
        v1.3.2
    """
    ver = parse_version()
    ver = ver.split('-')[0]

    release_type = RELEASE_MAP[_VERSION["release"]]
    release_num = _VERSION["release_num"]
    release_str = f" {release_type} {'' if _VERSION['release'].lower() == 'final' else f'({release_num})'}"
    return f'v{ver}{release_str}'


def parse_version() -> str:
    """
    Parses the version information into a string.

    Returns:
        str: The version information.

    Since:
        v1.3.2
    """
    version = f'{_VERSION["major"]}.{_VERSION["minor"]}.{_VERSION["patch"]}'

    if _VERSION['release'] != 'final':
        version += f'-{_VERSION["release"]}.{_VERSION["release_num"]}'

    return version


PYPI_VERSION_INFO = PyPiVersionInfo('inspyre-toolbox')
